/**********************************************************************
  Filename    : Camera Remote Bluetooth IoT
  Description : The esp32 controls remote android camera via bluetooth.
                This version is for the M5Stack Atom Lite (esp32 pico) controller.
  Function    : Photo, video recording, flash mode, and switch camera (v2)
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2022/02/09
***********************************************************************
//  M5Stack Atom Lite
//  https://docs.m5stack.com/en/core/atom_lite

//  Jpeg Decoder library for jpg image display
//  https://github.com/Bodmer/TJpg_Decoder

//  TFT_eSPI library for ST7789
//  For Atom Lite, need to update User_Setup.h code accordingly.
//  User_Setup.h file is located in the Arduino libraries folder.
//  C:\Users\UserName\Documents\Arduino\libraries\TFT_eSPI
//  https://github.com/Bodmer/TFT_eSPI
/**********************************************************************/
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"

#include <M5Atom.h>
#include "BluetoothSerial.h"
#include "string.h"
#include "SPI.h"
#include <TJpg_Decoder.h>
#include <TFT_eSPI.h> 

BluetoothSerial SerialBT;
TFT_eSPI tft = TFT_eSPI(); 

int APP_MODE = 5;             // 0: Switched Mode, 1: Joystick Mode, 5: M5
String dataValues[15];        // the data or status values from remote camera  
// Bluetooth Status
int CONNECTED = 0;
boolean DEBUG = false;

int ACTION_STATUS_NONE = 0;
int ACTION_STATUS_BTN_ACTION = 1;
int ACTION_STATUS_BTN_PRESSED = 2;
int ACTION_STATUS_BTN_LONG_PRESSED = 3;
int ACTION_STATUS_BTN_RELEASED = 4;

int ACTION_STATUS = ACTION_STATUS_NONE;

int FLASH_STATUS = 0;         //1: torch flash, 0: auto
int RECORDING_STATUS = 0;

int MODE_NONE = 0;
int MODE_PHOTO = 1;           //1: photo mode, 0: video mode
int MODE_VIDEO = 2;           //1: video mode, 0: photo mode
int MODE_VIDEO_RECORDING = 22; // recording video led flashing
int MODE_FLASH = 3;           //  1: torch flash, 0: auto
int MODE_CURRENT = MODE_PHOTO;

String actionString = "iot:";
String displayType = "ST7789";    // "none", "SD1306", "ST7789"
String displayMode = "0";         // text: "0", information: "1", image: "2"
String displayText = "Not Connected";     // PHOTO, VIDEO, REC..

#define PIN_BUTTON_ACTION         33      // action, take photo or record(stop) video 
#define PIN_BUTTON_TOGGLE_MODE    22 
int buttonExtraPressed = 0;

/***** TFT Display SPI PIN: See User_Setup.h ************************
//#define TFT_MISO 22   // no connection
//#define TFT_MOSI 19
//#define TFT_SCLK 23
//#define TFT_CS   33  // no connection Chip select control pin
//#define TFT_DC    21  // Data Command control pin
//#define TFT_RST   25  // Reset pin (could connect to RST pin)
*************************************************************/

// 192 x 128: image frame data buffer for displaying
//uint16_t image_data_color[24576];
byte* jpegData;
byte* image_data_receiving;               // temporary frame data buffer for remote image
bool imageDataProcessing = false;  

// memory allocation for image frame data receiving from remote camera
// memory is prepared on setup()
void prepareImageDataReceiving() {
  heap_caps_print_heap_info(MALLOC_CAP_DEFAULT);
  image_data_receiving = (byte*)malloc(12288 * sizeof(byte));

  if (!image_data_receiving) {
    Serial.println("failed for image_data_receiving malloc");
  } else {
    Serial.println("success for image_data_receiving malloc");
  }
}

// TFT_eSPI image decoder callback
// This next function will be called during decoding of the jpeg file to render each
bool tft_output(int16_t x, int16_t y, uint16_t w, uint16_t h, uint16_t* bitmap)
{
   // Stop further decoding as image is running off bottom of screen
  if ( y >= tft.height() ) return 0;

  tft.pushImage(x, y, w, h, bitmap);  // Blocking, so only returns when image block is drawn

  // Return 1 to decode next block.
  return 1;
}

void displayJpeg(int dataLength) {
  jpegData = (byte*)malloc(dataLength * sizeof(byte));

  for(int i = 0; i < dataLength; i++) {
    jpegData[i] = image_data_receiving[i];
  }

  // image size expected: width 192, height 128
  //TJpgDec.drawJpg(24, 24, jpegData, dataLength * sizeof(byte));
  // image size expected: width 240, height 160
  TJpgDec.drawJpg(0, 0, jpegData, dataLength * sizeof(byte));

  free(jpegData);
}

//  RGB Led display
uint8_t DisBuff[2 + 5 * 5 * 3]; //Used to store RBG color values.
void setBuff(uint8_t Rdata, uint8_t Gdata, uint8_t Bdata){  //Set the colors of LED, and save the relevant data to DisBuff[].
    DisBuff[0] = 0x05;
    DisBuff[1] = 0x05;
    for (int i = 0; i < 25; i++){
        DisBuff[2 + i * 3 + 0] = Rdata;
        DisBuff[2 + i * 3 + 1] = Gdata;
        DisBuff[2 + i * 3 + 2] = Bdata;
    }
}
void displayLed() {
  if (MODE_CURRENT == MODE_NONE) {
     setBuff(0x20, 0x20, 0x20);
  } else if (MODE_CURRENT == MODE_PHOTO) {
     setBuff(0x00, 0x40, 0x00);
  } else if (MODE_CURRENT == MODE_VIDEO) {
     setBuff(0x40, 0x00, 0x00);
  } else if (MODE_CURRENT == MODE_FLASH) {
     setBuff(0x00, 0x00, 0x40);
  }  
  M5.dis.displaybuff(DisBuff);    //Display the DisBuff color on the LED.    
}

//  Bluetooth event callback
void callback(esp_spp_cb_event_t event, esp_spp_cb_param_t *param) {
  if(event == ESP_SPP_SRV_OPEN_EVT){
    Serial.println("Client Connected");
    //digitalWrite(LED,HIGH);
    CONNECTED = true;

    MODE_CURRENT = MODE_PHOTO;
    displayMode = "2";
    displayLed();

    tft.fillScreen(TFT_BLACK);

    delay(100);
    sendDisplayTypeToRemote();
    delay(100);
    sendImageTypeToRemote();   
    delay(100);
    sendDisplayModeImageToRemote(); 
    delay(100);
    sendPreferredImageSize();
  }
 
  if(event == ESP_SPP_CLOSE_EVT ){
    Serial.println("Client disconnected");
    //digitalWrite(LED,LOW);
    CONNECTED = false;
    
    MODE_CURRENT = MODE_NONE;
    displayMode = "0";
    displayLed();
    displayLogo();
  }
}

//  Display Logo (or, text)
void displayLogo() {
  tft.fillScreen(TFT_BLACK);
// Set "cursor" at top left corner of display (0,0) and select font 2 
  tft.setCursor(40, 40, 4);
// Set the font colour to be white with a black background
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.println("CAMERA\n");
  tft.setCursor(40, 100, 4);
  tft.println("REMOTE\n");
  tft.setCursor(40, 160, 4);
  tft.println("BLUETOOTH\n");  
}


void setup() {

  MODE_CURRENT = MODE_NONE;

//  init M5Atom 
//  void begin(bool SerialEnable, bool I2CEnable, bool DisplayEnable);
  M5.begin(true, false, true);
  
  delay(10);
  displayLed();
  
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0); //disable brownout detector   ds

  SerialBT.begin("Camera Remote Bluetooth"); //Bluetooth device name
  SerialBT.register_callback(callback);
  
  Serial.println("\nThe device started, now you can pair it with bluetooth!");
  
  delay(500);
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 1); //enable brownout detector

//  Initialise the TFT Display
  tft.begin();
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.fillScreen(TFT_BLACK);
//  The jpeg image can be scaled down by a factor of 1, 2, 4, or 8
  TJpgDec.setJpgScale(1);
// The colour byte order can be swapped by the decoder
// using TJpgDec.setSwapBytes(true); or by the TFT_eSPI library:
  tft.setSwapBytes(true);  
//  The decoder must be given the exact name of the rendering function above
  TJpgDec.setCallback(tft_output);  

//  Display Logo (or, text)
  displayLogo();

  // allocate memory for the remote image frame data
  prepareImageDataReceiving();

  pinMode(PIN_BUTTON_ACTION, INPUT_PULLUP);
  pinMode(PIN_BUTTON_TOGGLE_MODE, INPUT_PULLUP);
  
  delay(500);
}

void processAction() {
  //You need to add m5.update () to read the state of the key, see System for details
  M5.update();
  
//int ACTION_STATUS_NONE = 0;
//int ACTION_STATUS_BTN_ACTION = 1;
//int ACTION_STATUS_BTN_PRESSED = 2;
//int ACTION_STATUS_BTN_LONE_PRESSED = 3;
//int ACTION_STATUS_BTN_RELEASED = 4;

  int action = digitalRead(PIN_BUTTON_ACTION);
  int camera_mode = digitalRead(PIN_BUTTON_TOGGLE_MODE);

  if (!action || !camera_mode) {
    String whichButton = (!action ? "action" : "camera_mode");
    if (ACTION_STATUS == ACTION_STATUS_NONE) {
      Serial.println("Button pressed: " + whichButton);
      ACTION_STATUS = ACTION_STATUS_BTN_PRESSED; 
      if (!action) {
        buttonExtraPressed = 1;
      } else {
        buttonExtraPressed = 2;
      }
    }    
    return;   
  }
  if (ACTION_STATUS == ACTION_STATUS_BTN_PRESSED) {

    if (buttonExtraPressed == 1) {
      Serial.println("Action button released.");
      // do remote action
      doAction();
      ACTION_STATUS = ACTION_STATUS_NONE;
      buttonExtraPressed = 0;
      return;
    } else if (buttonExtraPressed == 2) {
      if (MODE_CURRENT == MODE_PHOTO) {
        MODE_CURRENT = MODE_VIDEO;
      } else if (MODE_CURRENT == MODE_VIDEO) {
        MODE_CURRENT = MODE_PHOTO;
      } 
      Serial.println("Toggle mode button released: " + String(MODE_CURRENT));
      displayLed();
      doModeAction();
      ACTION_STATUS = ACTION_STATUS_NONE;   
      buttonExtraPressed = 0; 
      return;
    }   
  }

  if (ACTION_STATUS == ACTION_STATUS_NONE || ACTION_STATUS == ACTION_STATUS_BTN_PRESSED) {

    if (M5.Btn.pressedFor(2000)) {
      // If the button is pressed for more than 2 seconds
      // switch camera: front or back (send to remote)
      Serial.println("Button long pressed: for switch camera");
      
      //tft.fillScreen(TFT_BLACK);
      ACTION_STATUS = ACTION_STATUS_BTN_LONG_PRESSED;

      tft.fillRect(0, 200, 240, 40, TFT_BLACK);
      tft.setCursor(40, 200, 4);
      tft.setTextColor(TFT_WHITE, TFT_BLACK);
      tft.println("Switch Camera");      
            
    } else {

      if (ACTION_STATUS == ACTION_STATUS_NONE) {
        if (M5.Btn.isPressed()) {
          Serial.println("Button pressed for flash");
          ACTION_STATUS = ACTION_STATUS_BTN_PRESSED;

          tft.fillRect(0, 200, 240, 40, TFT_BLACK);
          tft.setCursor(40, 200, 4);
          tft.setTextColor(TFT_WHITE, TFT_BLACK);
          tft.println("Flash Mode");           
        }
      }  else if (ACTION_STATUS == ACTION_STATUS_BTN_PRESSED) {
        if (M5.Btn.isReleased()) {
          Serial.println("Button released for flash.");
          
          doFlashAction();
          ACTION_STATUS = ACTION_STATUS_NONE;
        }               
      }
    }

  } else if (ACTION_STATUS == ACTION_STATUS_BTN_LONG_PRESSED) {
    // switch camera: front or back (send to remote)
    if (M5.Btn.isReleased()) {
      Serial.println("Button released for switch camera.");
      
      displayLed();
      doSwitchCameraAction();
      ACTION_STATUS = ACTION_STATUS_NONE;
    }         
  }

}

void doSwitchCameraAction() {
  String photo = (MODE_CURRENT == MODE_PHOTO ? "1" : "0");
  String video = (MODE_CURRENT == MODE_VIDEO ? "1" : "0");
  actionString = "iot:";
  tft.fillScreen(TFT_BLACK);
  actionString += photo + ":" + video + ":0:0:1:0:" + displayType;    
  if (CONNECTED) {
    Serial.println(actionString);
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   

    delay(250);
  } 

  if (MODE_CURRENT == MODE_PHOTO) {
    displayText = "PHOTO MODE";
  } else if (MODE_CURRENT == MODE_VIDEO) {
    displayText = "VIDEO MODE";
  }
  tft.fillRect(0, 200, 240, 40, TFT_BLACK);
  tft.setCursor(40, 200, 4);
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.println(displayText);      
      
}

void doFlashAction() {
  actionString = "iot:";
  FLASH_STATUS = (FLASH_STATUS == 1 ? 0 : 1);
  actionString += "0:0:" + String(FLASH_STATUS) + ":0:0:0:" + displayType;    
  if (CONNECTED) {
    Serial.println(actionString);
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   

    delay(250);
  } 

  if (MODE_CURRENT == MODE_PHOTO) {
    displayText = "PHOTO MODE";
  } else if (MODE_CURRENT == MODE_VIDEO) {
    displayText = "VIDEO MODE";
  }
  tft.fillRect(0, 200, 240, 40, TFT_BLACK);
  tft.setCursor(40, 200, 4);
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.println(displayText);      
         
}

void doAction() {
  
//int MODE_NONE = 0;
//int MODE_PHOTO = 1;           //1: photo mode, 0: video mode
//int MODE_VIDEO = 2;           //1: video mode, 0: photo mode
//int MODE_VIDEO_RECORDING = 22; // recording video led flashing
//int MODE_FLASH = 3;   

  actionString = "iot:";
  
  if (MODE_CURRENT == MODE_PHOTO) {
    displayText = "PHOTO MODE";
    actionString += "0:0:0:1:0:0:" + displayType;
  } else if (MODE_CURRENT == MODE_VIDEO) {
    displayText = "VIDEO MODE";
    actionString += "0:0:0:1:0:0:" + displayType;
  } else if (MODE_CURRENT == MODE_VIDEO_RECORDING) {
    displayText = "VIDEO MODE";
    actionString += "0:0:0:1:0:0:" + displayType;
  } else if (MODE_CURRENT == MODE_FLASH) {
//    FLASH_STATUS = (FLASH_STATUS == 1 ? 0 : 1);
//    actionString += "0:0:" + String(FLASH_STATUS) + ":0:0:0:" + displayType;
  }
  Serial.println("doAction: " + actionString);

  if (CONNECTED) {
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   

    delay(250);
  }

  tft.fillScreen(TFT_BLACK);
  //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
  tft.setCursor(40, 200, 4);
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.println(displayText);   
}

void doModeAction() {
  
//int MODE_NONE = 0;
//int MODE_PHOTO = 1;           //1: photo mode, 0: video mode
//int MODE_VIDEO = 2;           //1: video mode, 0: photo mode
//int MODE_VIDEO_RECORDING = 22; // recording video led flashing
//int MODE_FLASH = 3;   

  actionString = "iot:";

  if (MODE_CURRENT == MODE_PHOTO) {
    actionString += "1:0:0:0:0:0:" + displayType;
    displayText = "PHOTO MODE";
  } else if (MODE_CURRENT == MODE_VIDEO) {
    actionString += "0:1:0:0:0:0:" + displayType;
    displayText = "VIDEO MODE";
  } else if (MODE_CURRENT == MODE_FLASH) {
  } 
  Serial.println(actionString);

  if (CONNECTED) {
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   

    delay(250);
  }

  tft.fillScreen(TFT_BLACK);
  //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
  tft.setCursor(40, 200, 4);
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.println(displayText); 
}

// send Jpeg decode mode to remote camera
// remote camera device will send preview frame as jpeg format
void sendImageTypeToRemote() {
  if (CONNECTED) {
    SerialBT.println("imagetype:jpeg");
    SerialBT.write(13);
    SerialBT.write(10);    
  }
}
// send current display type to remote camera
// remote camera device will sets proper image frame data and status information for transmision
void sendDisplayTypeToRemote() {
  if (CONNECTED) {
    SerialBT.println("displaytype:ST7789");
    SerialBT.write(13);
    SerialBT.write(10);
  }
}

// force to use image mode to the remote camera
void sendDisplayModeImageToRemote() {
  if (CONNECTED) {
    actionString = "iot:0:0:0:0:0:1:" + displayType;
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   
  }  
}

// send preferred image frame size by screen width and height
void sendPreferredImageSize() {
  if (CONNECTED) {
    actionString = "imagesize:240:160:";
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   
  }
}

void clearDataValues() {
  int numberOfValues = 10;    //(int)sizeof(dataValues)/sizeof("hello");
  for (int i = 0; i < numberOfValues; i++) {
    dataValues[i] = "";
  }
}

// parse the remote data headers and values
// split data string into array
// delimeter: ":"
void parseData(String dataString, boolean processData) {
  if (dataString.length() < 4) {
    return;
  }
  clearDataValues();

  // data delimeter: ":"
  char delimeter = ':';
  int findStart = 0;
  int findIndex = 0;
  int findCount = 0;

  findIndex = dataString.indexOf(delimeter, findStart);
  if (findIndex > 0) {
    dataValues[findCount] = dataString.substring(findStart, findIndex);
    findCount ++;
    findStart = findIndex + 1;

    while (findIndex > 0) {
      findIndex = dataString.indexOf(delimeter, findStart);
      if (findIndex > 0) {
        dataValues[findCount] = dataString.substring(findStart, findIndex);
        findCount ++;
        findStart = findIndex + 1;        
      }
    } 
  }

  if (!processData) {
    // image data header parsing.
    return;
  }
  if (imageDataProcessing) {
    // error receiving the frame data, not synced or serialized
    delay(500);
    return;
  }
  if (DEBUG) {
    Serial.println("parseData: " + dataString);
  }

  String dataName = dataValues[0];

  if (dataName.equalsIgnoreCase("received")) {
    String statusPhoto = dataValues[1];
    int photoValue = statusPhoto.toInt();
    String statusVideo = dataValues[2];
    int videoValue = statusVideo.toInt();

    if (photoValue == 1) {
      if (MODE_CURRENT != MODE_PHOTO) {
      }
      MODE_CURRENT = MODE_PHOTO;
      displayLed();
      displayText = "PHOTO MODE";
      tft.fillScreen(TFT_BLACK);
      //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
      tft.setCursor(40, 200, 4);
      tft.setTextColor(TFT_WHITE, TFT_BLACK);
      tft.println(displayText);           
    } else if (videoValue == 1) {
      if (MODE_CURRENT != MODE_VIDEO) {
      }
      MODE_CURRENT = MODE_VIDEO;
      displayLed();
      displayText = "VIDEO MODE";
      tft.fillScreen(TFT_BLACK);
      //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
      tft.setCursor(40, 200, 4);
      tft.setTextColor(TFT_WHITE, TFT_BLACK);
      tft.println(displayText);           
    }   
           
  } else if (dataName.equalsIgnoreCase("iot")) {
    // current remote camera status
    // iot:0:1:0:0:ok:0:0:192:
    Serial.println("parseData: " + dataString);
    
    String statusPhoto = dataValues[1];
    int photoValue = statusPhoto.toInt();
    String statusVideo = dataValues[2];
    int videoValue = statusVideo.toInt();
    String statusFlash = dataValues[3];
    int flashValue = statusFlash.toInt();
    String statusAction = dataValues[4];
    int actionValue = statusAction.toInt();
    String statusResult = dataValues[5];
    int resultValue = statusResult.toInt();
    displayMode = dataValues[6];  // none: 0, text: 1, image: 2
    String previewSize = dataValues[7];
    String pixelsOnData = dataValues[8];   

    FLASH_STATUS = flashValue;

    if (MODE_CURRENT != MODE_FLASH) {
      if (photoValue == 1) {
        if (MODE_CURRENT != MODE_PHOTO) {
          MODE_CURRENT = MODE_PHOTO;
          displayLed();
          displayText = "PHOTO MODE";
          tft.fillScreen(TFT_BLACK);
          //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
          tft.setCursor(40, 200, 4);
          tft.setTextColor(TFT_WHITE, TFT_BLACK);
          tft.println(displayText);     
        }
      } else if (videoValue == 1) {
        if (MODE_CURRENT != MODE_VIDEO) {
          MODE_CURRENT = MODE_VIDEO;
          displayLed();
          displayText = "VIDEO MODE";
          tft.fillScreen(TFT_BLACK);
          //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
          tft.setCursor(40, 200, 4);
          tft.setTextColor(TFT_WHITE, TFT_BLACK);
          tft.println(displayText);     
        }
      }      
    }

    if (photoValue == 1 && !displayText.equals("PHOTO MODE")) {
        displayText = "PHOTO MODE";    
        tft.fillScreen(TFT_BLACK);
        //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
        tft.setCursor(40, 200, 4);
        tft.setTextColor(TFT_WHITE, TFT_BLACK);
        tft.println(displayText);         
    } else if (videoValue == 1 && !displayText.equals("VIDEO MODE")) {
        displayText = "VIDEO MODE";
        tft.fillScreen(TFT_BLACK);
        //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
        tft.setCursor(40, 200, 4);
        tft.setTextColor(TFT_WHITE, TFT_BLACK);
        tft.println(displayText);     
    }

    // force to use image mode for this version
    if (!displayMode.equals("2")) {
      sendDisplayModeImageToRemote();
    }

  } else if (dataName.equalsIgnoreCase("photo")) {
    if (MODE_CURRENT != MODE_PHOTO) {
      MODE_CURRENT = MODE_PHOTO;
      displayLed();
    } else {
      displayText = "PHOTO MODE";    
      //tft.fillScreen(TFT_BLACK);
      tft.fillRect(0, 180, 240, 60, TFT_BLACK);
      tft.setCursor(40, 200, 4);
      tft.setTextColor(TFT_WHITE, TFT_BLACK);
      tft.println(displayText);   
    }
  } else if (dataName.equalsIgnoreCase("video")) {
    // video recording status
    // video:2:, video:1:, ...
    String recordingStatus = dataValues[1];
    
    if (MODE_CURRENT != MODE_VIDEO) {
      MODE_CURRENT = MODE_VIDEO;
      displayLed();
    } else {
      // show video recording status
      // blink led

      if (recordingStatus.equals("2")) {
        // video:2: recording
        if (RECORDING_STATUS == 0) {
          setBuff(0x10, 0x00, 0x00);
          M5.dis.displaybuff(DisBuff);    //Display the DisBuff color on the LED.  
          RECORDING_STATUS = 1;
  
          displayText = "REC..";
          tft.fillRect(0, 180, 240, 60, TFT_BLACK);
          tft.setCursor(80, 200, 4);
          tft.setTextColor(TFT_WHITE, TFT_BLACK);
          tft.println(displayText);        
        } else if (RECORDING_STATUS == 1) {
          setBuff(0x40, 0x00, 0x00);
          M5.dis.displaybuff(DisBuff);    //Display the DisBuff color on the LED. 
          RECORDING_STATUS = 0; 
  
          displayText = "REC.";
          tft.fillRect(0, 180, 240, 60, TFT_BLACK);
          tft.setCursor(80, 200, 4);
          tft.setTextColor(TFT_WHITE, TFT_BLACK);
          tft.println(displayText);        
        }         
      } else {
        // video:1: stopped
          displayText = "VIDEO MODE";
          tft.fillRect(0, 180, 240, 60, TFT_BLACK);
          tft.setCursor(40, 200, 4);
          tft.setTextColor(TFT_WHITE, TFT_BLACK);
          tft.println(displayText);        
      }

      displayText = "VIDEO MODE";
    }
    
  }
  
}

void processImageData(String header) {
  // flag data processing, prevent another incoming data processing
  // until this image display session finished
  imageDataProcessing = true;

  if (header.startsWith("imagedata")) {
    parseData(header, false);
    int photoMode = dataValues[3].toInt();
    int videoMode = dataValues[4].toInt();

    if (photoMode == 1) {
      if (MODE_CURRENT != MODE_PHOTO && (ACTION_STATUS == ACTION_STATUS_NONE)) {
        MODE_CURRENT = MODE_PHOTO;
        displayText = "PHOTO MODE";
        displayLed();
        tft.fillScreen(TFT_BLACK);
        //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
        tft.setCursor(40, 200, 4);
        tft.setTextColor(TFT_WHITE, TFT_BLACK);
        tft.println(displayText);            
      }
    } else if (videoMode == 1) {
      if (MODE_CURRENT != MODE_VIDEO && (ACTION_STATUS == ACTION_STATUS_NONE)) {
        MODE_CURRENT = MODE_VIDEO;
        displayText = "VIDEO MODE";
        displayLed();
        tft.fillScreen(TFT_BLACK);
        //tft.fillRect(0, 180, 240, 60, TFT_BLACK);
        tft.setCursor(40, 200, 4);
        tft.setTextColor(TFT_WHITE, TFT_BLACK);
        tft.println(displayText);          
      } 
    }
  }

  while (imageDataProcessing) {
    if (DEBUG) {
      Serial.println("processImageData: " + header);  
    }    
    // image frame data information header to receive into memory
    String pixelsDataInfo = SerialBT.readStringUntil('\n');
    parseData(pixelsDataInfo, false);
    
    String dataName = dataValues[0];
    String colorMode = dataValues[4];
    int totalPixels = dataValues[2].toInt();    

    if (dataName.startsWith("imagepixels_color")) {
      if (DEBUG) {
        Serial.println("processImageData: " + pixelsDataInfo);   
      }    

      // this version only uses jpeg image format
      if (colorMode.startsWith("jpeg")) {
  
        int count = 0;
        while (count < totalPixels) {
          count += SerialBT.readBytes(image_data_receiving, totalPixels);
        }  
  
        if (count <= 0) {
          delay(100);
        } else {
          displayJpeg(totalPixels);        
        }
        imageDataProcessing = false;          
      }
          
    }

    processAction();
  }
  
}

void loop() {

  if (CONNECTED) {
    
    while(SerialBT.available())
    {
      
      if (!imageDataProcessing) {
        String received = SerialBT.readStringUntil('\n');

        if (received.startsWith("imagedata")) {
          imageDataProcessing = true;
          processImageData(received);
        } else {
          parseData(received, true);         
        }
        
      }

    }

    processAction();

  } else { //end of: if (CONNECTED)

    // not connected
    //displayLogo();
  }

  delay(10);
}
